pub use {
    super::B64Alphabet,
    std::io::{self, ErrorKind, Read, Write},
};

#[derive(Debug)]
pub enum Error {
    Io(io::Error),
    IllegalChar(char),
    MissingPadding,
}

impl From<io::Error> for Error {
    fn from(value: io::Error) -> Self {
        Self::Io(value)
    }
}

impl From<char> for Error {
    fn from(value: char) -> Self {
        Self::IllegalChar(value)
    }
}

pub struct Decoder<R: Read, W: Write> {
    reader: R,
    writer: W,
    alphabet: B64Alphabet,
}

impl<R: Read, W: Write> Decoder<R, W> {
    pub fn new(reader: R, writer: W, alphabet: Option<B64Alphabet>) -> Self {
        Self {
            reader,
            writer,
            alphabet: alphabet.unwrap_or_default(),
        }
    }

    pub fn decode(&mut self) -> Result<(), Error> {
        loop {
            let mut ibuf = [0_u8; 4];
            let mut obuf = [0_u8; 3];
            let mut num = 0_u32;
            let mut n_bytes = 0;
            loop {
                n_bytes += match self.reader.read(&mut ibuf) {
                    Ok(n) => n,
                    Err(e) if e.kind() == ErrorKind::Interrupted => continue,
                    Err(e) => return Err(e.into()),
                };
                break;
            }
            match n_bytes {
                0 => break,
                4 => {}
                _ => return Err(Error::MissingPadding),
            }
            let mut bytes = 0;
            for (i, &c) in ibuf.iter().enumerate() {
                let c = c.into();
                num <<= 6;
                if c == self.alphabet.pad() {
                    continue;
                }
                if i != bytes {
                    return Err(Error::IllegalChar(c));
                }
                let Some(idx) = self.alphabet.get(c) else {
                    return Err(Error::IllegalChar(c));
                };
                num |= idx as u32;
                bytes += 1;
            }
            let olen = bytes * 6 / 8;
            for i in (0..3).rev() {
                obuf[i] = (num & 0xff) as u8;
                num >>= 8;
            }
            self.writer.write_all(&mut obuf[0..olen])?;
        }
        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn decode() {
        let mut decoder = Decoder::new("SGVsbG8sIFdvcmxk".as_bytes(), vec![], None);
        decoder.decode().unwrap();
        assert_eq!(String::from_utf8(decoder.writer).unwrap(), "Hello, World");
        decoder = Decoder::new("SGVsbG8sIFdvcmxkIQ==".as_bytes(), vec![], None);
        decoder.decode().unwrap();
        assert_eq!(String::from_utf8(decoder.writer).unwrap(), "Hello, World!");
        decoder = Decoder::new("SGVsbG8sIFdvcmxkIQo=".as_bytes(), vec![], None);
        decoder.decode().unwrap();
        assert_eq!(
            String::from_utf8(decoder.writer).unwrap(),
            "Hello, World!\n"
        );
    }
}
